/***********************************************************************
 * "seq" files template
 *
 * This header works with "linux/seq_file.h"
 *
 * Licenced under GPL.
 *
 * Copyright (C) YOKOTA Hiroshi <yokota (at) netlab. is. tsukuba. ac. jp>
 */

#ifndef _LINUX_SEQ_TEMPLATE_H_
#define _LINUX_SEQ_TEMPLATE_H_

/************************************************************************
 * Internal use only.
 */

/* "seq" initializer */
#define _SEQ_OPEN_FS(_name_) \
static int _name_ ## _open_fs(struct inode *inode, struct file *file) \
{								      \
	return single_open(file, _name_ ## _show, PDE(inode)->data);  \
}

/* "seq" reader */
#define _SEQ_READER(_name_) \
static int _name_ ## _show(struct seq_file *seq, void *offset);

/* "seq" writer */
#define _SEQ_WRITER(_name_) \
static ssize_t _name_ ## _write(struct file	  *file,   \
				const char __user *buffer, \
				size_t		   count,  \
				loff_t		  *ppos);

/*************************************************************************
 * "seq" file template definition.
 */

/*-------------------------------------------------------------------------
 * "seq" files template for read-only files.
 * You must add your own [_name_]_show(struct seq_file *, void *).
 */
#define SEQ_FILEOPS_TEMPLATE_R(_name_) \
 \
_SEQ_READER(_name_)  \
_SEQ_OPEN_FS(_name_) \
 \
static struct file_operations _name_ ## _fops = { \
	.open	 = _name_ ## _open_fs,		  \
	.read	 = seq_read,			  \
	.llseek	 = seq_lseek,			  \
	.release = single_release,		  \
}

/*------------------------------------------------------------------------
 * "seq" files template for read-write files.
 * You must add your own [_name_]_show(struct seq_file *, void *) and
 * [_name_]_write(struct file*, const char __user *, size_t, loff_t *).
 */
#define SEQ_FILEOPS_TEMPLATE_RW(_name_) \
 \
_SEQ_READER(_name_)  \
_SEQ_WRITER(_name_)  \
_SEQ_OPEN_FS(_name_) \
 \
static struct file_operations _name_ ## _fops = { \
	.open	 = _name_ ## _open_fs,		  \
	.read	 = seq_read,			  \
	.write	 = _name_ ## _write,		  \
	.llseek	 = seq_lseek,			  \
	.release = single_release,		  \
}

#endif
/* end */
